/*
   Try to use two explicit mutual exclusion (mutex) locks.
*/

import java.util.concurrent.locks.ReentrantLock;

public class Race_3_unsynchronized
{
   // "global" variables
   private static ReentrantLock mutex1 = new ReentrantLock();
   private static ReentrantLock mutex2 = new ReentrantLock();
   private static boolean runFlag = true;  // No need for volatile keyword.
   private static long count  = 0;  // This variable is shared by thread1 and thread2.
   private static long count1 = 0;  // This variable is used only by thread1.
   private static long count2 = 0;  // This variable is used only by thread2.

   //public static Object myStupidLock = new Object();

   public static void main(String[] args)
   {
      Function1 f1 = new Function1();
      Function2 f2 = new Function2();

      // Create two threads.
      System.out.println("Race_3_unsynchronized: Starting threads");
      Thread t1 = new Thread( f1 );
      Thread t2 = new Thread( f2 );

      t1.start();
      t2.start();

      try
      {
         Thread.sleep(2000);  // Let the threads run for a while.

         runFlag = false; // Signal the threads to stop.

         t1.join();
         t2.join();

         System.out.println("\nThreads stopped.");
         System.out.printf("count = %,d  count1 = %,d  count2 = %,d\n", count, count1, count2);
      }
      catch (InterruptedException e)
      {
         e.getMessage();
      }
   }


   private static class Function1 implements Runnable
   {
      synchronized public void run() // This is NOT "synchronized". Why?
      {
         long new_count, old_count;
         while (runFlag)
         {
            mutex1.lock();  // This is NOT "synchronized". Why?
              old_count = count++;   // 1. read count, 2. write old_count, 3. inc count
              new_count = count;     // 4. read count, 5. write new_count
            mutex1.unlock();
            // Check that the local and global variables were consistent with each other.
            if (new_count - old_count != 1)
               System.out.printf("Bingo1: new_count - old_count = %d\n", new_count-old_count);
            count1++;
         }
      }
   }


   private static class Function2 implements Runnable
   {
      synchronized public void run() // This is NOT "synchronized". Why?
      {
         long new_count, old_count;
         while (runFlag)
         {
            mutex2.lock();  // This is NOT "synchronized". Why?
              old_count = count++;   // 1. read count, 2. write old_count, 3. inc count
              new_count = count;     // 4. read count, 5. write new_count
            mutex2.unlock();
            // Check for the race condition.
            if (new_count - old_count != 1)
               System.out.printf("Bingo2: new_count - old_count = %d\n", new_count-old_count);
            count2++;
         }
      }
   }
}
